function DrawingManager(map, opts) {

    try {
        ymaps;
    } catch (e) {
        throw Error('Ymaps Map JS API is not ready yet!');
    }

    let drawModeMarker = "marker";
    let DrawModeCircle = "circle";
    let DrawModeArea = "area";
    let DrawModeFitBounds = "fitbounds";
    let DrawModeReset = "reset";

    let me = this;
    me.map = map;
    me._opts = opts;
    me._drawingType = null;
    me.markerOptions = opts.markerOptions || {};
    me.circleOptions = opts.circleOptions || {};
    me.areaOptions = opts.areaOptions || {};
    me.isDraw = opts.enableDraw;
    me.radius = opts.circleOptions.radius;
    me.vertex = [];
    me.vertexCount = 0;
    me.polygonArea = new Map();
    me.isMoreArea = opts.areaOptions.isMoreArea || false;
    me.squareArea = opts.areaOptions.areaMaxSquare || 1400;
    me.isRedraw = null;
    me.mode = opts.activeMode;

    let vertexPane = new ymaps.pane.MovablePane(me.map, { zIndex: 4000 });
    me.map.panes.append('vertex', vertexPane);

    DrawingManager.prototype.setDrawingMode = function (drawingType, opts) {
        const me = this;

        me.markerOptions = opts.markerOptions || {};
        me.circleOptions = opts.circleOptions || {};
        me.areaOptions = opts.areaOptions || {};
        me.isDraw = opts.isDraw;
        me.isRedraw = opts.isRedraw;
        me.mode = opts.activeMode;

        const eventMode = {
            isRedraw: opts.isRedraw,
            isDraw: opts.isDraw,
            mode: me.mode,
        }

        me.map.events.fire('draw:mode', { eventMode });

        if (me.markerOptions) {
            me.setPosition(me.markerOptions.markerLatitude, me.markerOptions.markerLongitude);
        }
        if (me.circleOptions) {
            me.radius = me.circleOptions.radius;
        }
        if (me.areaOptions) {
            me.areaCoordinates = me.areaOptions.areaCoordinates;
        }

        switch (drawingType) {
            case drawModeMarker:
                me._bindMarker(false);
                break;
            case DrawModeCircle:
                me._bindMarker(true);
                break;
            case DrawModeArea:
                me._bindArea();
                break;
            case DrawModeFitBounds:
                me._fitBounds();
            case DrawModeReset:
                me.reset();
                break;
            default:
                me._remove();
                break;
        }
        me._drawingType = drawingType;
    };

    DrawingManager.prototype.reset = function () {
        const me = this;
        me.map.container.events.remove('click');
        me.map.container.events.remove('mousedown');
        me.map.container.events.types.click = undefined;
        me.isRedraw = false;

    };

    DrawingManager.prototype.remove = function () {
        const me = this;
        me._remove();
        me.isRedraw = false;
    };

    DrawingManager.prototype._remove = function () {
        const me = this;
        if (me._area || me.polygonArea.length) { me._removeArea(); }
        if (me._circle) { me._removeCircle(); }
        if (me._centerMarker) { me._removeCenterMarker(); }
        me.map.container.events.remove('click');
        me.map.container.events.remove('mousedown');
        me.map.container.events.types.click = undefined;
    },

        DrawingManager.prototype.resetMarker = function () {
            const me = this;
            me._centerMarker && me._removeCenterMarker();
        },

        DrawingManager.prototype.isAnimateMarker = function () {
            try {
                const me = this;
                markerElement = me._centerMarker.getOverlaySync().getLayoutSync().getElement();
                markerElement.classList.add('poi', 'drawing');
            } catch { }
        },

        DrawingManager.prototype._fitBounds = function () {
            const me = this;
            me._centerMarker && me._fitBoundsMarker();
            me._area && me._fitBoundsArea();
            me._circle && me._fitBoundsCircle();
        };

    DrawingManager.prototype.setPosition = function (latitude, longitude) {
        const me = this;
        me.position = null;
        me.position = [latitude, longitude];
    };

    DrawingManager.prototype._setPosition = function (e) {
        const me = this;
        me.position = null;
        me.position = e.get('coords');

    };

    DrawingManager.prototype.setEnableDraw = function (enabled) {
        this.isDraw = enabled;
    };

    DrawingManager.prototype._bindMarker = function (enabledCircle) {
        const me = this;
        me._removeCircle();
        me._removeArea();
        me._removeCenterMarker();

        me.map.container.events.remove('click');
        me.map.container.events.remove('mousedown');
        me.map.container.events.remove('mouseup');

        let createCenterMarker = (e) => {
            me._removeCircle();
            me._removeCenterMarker();

            e && me._setPosition(e);

            if (me.position) {
                const geo = {
                    geometry: {
                        type: "Point",
                        coordinates: me.position
                    }
                }

                const options = {
                    iconLayout: 'default#image',
                    iconImageSize: [24, 30],
                    iconImageOffset: [-10, -24],
                    syncOverlayInit: true,
                    iconImageHref: enabledCircle ? me.circleOptions.iconUrl : me.markerOptions.iconUrl,
                    draggable: false,
                }

                me._centerMarker = new ymaps.GeoObject(geo, options);
                me.map.geoObjects.add(me._centerMarker);

                !enabledCircle && me.map.events.fire('draw:marker_create', me._getPosition());
                me.markerOptions.fitBounds && !enabledCircle && me._fitBoundsMarker();

                enabledCircle && me._bindCircle();
                me.position = null;
                me.markerOptions.isMarkerAnimate && !enabledCircle && me.isAnimateMarker(true);

                me._centerMarker.events.add('click', (event) => {
                    if (!me._circle) {
                        me.map.events.fire('draw:marker_click', me._getPosition());
                        if (event && event.stopPropagation) {
                            event.stopPropagation();
                        } else {
                            window.event.cancelBubble = true;
                        }
                    }
                });



                me._centerMarker.events.add('mouseenter', (event) => {
                    const position = {
                        latitude: me._centerMarker.geometry.getCoordinates()[0],
                        longitude: me._centerMarker.geometry.getCoordinates()[1]
                    }

                    const projection = me.map.options.get('projection', {});
                    const pixel = me.map.converter.globalToPage(
                        projection.toGlobalPixels(
                            me._centerMarker.geometry.getCoordinates(),
                            me.map.getZoom()
                        )
                    );

                    const offset = me.map.container.getOffset();
                    const eventMarker = {
                        clientX: pixel[0] - offset[0],
                        clientY: pixel[1] - offset[1],
                        position
                    }

                    me.map.events.fire('draw:marker_mouseover', { eventMarker });
                });

                me._centerMarker.events.add('mouseleave', () => {
                    me.map.events.fire('draw:marker_mouseout', null);
                });

            }
        }

        if (!me.isDraw) {
            createCenterMarker();
            me.map.container.events.remove('click');

        }
        me.map.container.events.add('click', (event) => {
            if (event.isMapEventAllowed()) {
                if (this.isDraw || me.isRedraw) { createCenterMarker(event); }
            }
        });

    };

    DrawingManager.prototype._bindCircle = function () {
        const me = this;
        me._removeCircle();

        if (me._centerMarker) {
            const position = [me._centerMarker.geometry.getCoordinates(), Math.floor(me.radius)];
            const options = {
                fillColor: me.circleOptions.fillColor,
                fillOpacity: me.circleOptions.fillOpacity,
                strokeColor: me.circleOptions.strokeColor,
                strokeOpacity: me.circleOptions.strokeOpacity,
                strokeWidth: me.circleOptions.strokeWeight,
                syncOverlayInit: false,
                zIndex: 100,
                zIndexHover: 100,

            }

            me._circle = new ymaps.Circle(position, {}, options);
            me.map.geoObjects.add(me._circle);

            me._circle.events.add('click', function (event) { event.stopPropagation() });

            me._fitBoundsCircle();

            me._centerMarker.options.set({
                draggable: true,
                cursor: 'move',
                zIndex: 9999,
            });

            me._setIconMarker(me.circleOptions.iconUrl);
            me._centerMarkerAddEventListener();
            me._createVertexMarker();
            me._createVertexMarkerCopy();
        }
    };

    DrawingManager.prototype._bindArea = function () {
        let me = this;
        me.map.container.events.remove('click');
        me.map.container.events.types.click = undefined;
        !me.isMoreArea && me._removeArea();

        me._removeCircle();
        me._removeCenterMarker();

        if (me.isMoreArea && me.polygonArea.size === 5) { return false }
        let createArea = () => {

            let patch = [];
            let _patch = [];
            me._area = null;
            me._setDrawing(false);

            const options = {
                fillColor: me.areaOptions.fillColor,
                fillOpacity: me.areaOptions.fillOpacity,
                strokeColor: me.areaOptions.strokeColor,
                strokeOpacity: me.areaOptions.strokeOpacity,
                strokeWidth: me.areaOptions.strokeWeight,
                syncOverlayInit: true,
            }

            me._area = new ymaps.Polyline([], {}, options);
            me.map.geoObjects.add(me._area);
            me.map.container.events.add('mousemove', (event) => {
                if (typeof event.get('coords') !== "undefined") {
                    me.map.container.getElement().style.cursor = 'not-allowed';
                    patch.push(event.get('coords'));
                    const area = me.geodesicArea(patch);
                    if (area < me.squareArea * 1000000) {
                        _patch.push(event.get('coords'));
                        me._area.geometry.setCoordinates(_patch);
                        me.map.container.getElement().style.cursor = 'crosshair';
                        return false;
                    }
                }
            });

            document.addEventListener('mouseup', () => {
                me.map.container.events.remove('mousemove');
                me.map.container.events.types.mousemove = undefined;
                me.map.geoObjects.remove(me._area);
                me._setDrawing(true);

                if (!_patch.length) { return false };

                me._area = new ymaps.Polygon([_patch,], {}, options);
                me.map.geoObjects.add(me._area);
                me.vertexCount++;
                me.map.container.events.remove('mousedown');
                me.map.container.events.types.mousedown = undefined;
                me.polygonArea.set(me.vertexCount, me._area);

                me._fitBoundsArea(_patch);

                if (me.isMoreArea) {
                    patchAreaLatLng = _patch;
                    let patchAreaLatLngMax = me.findRoundPoints(patchAreaLatLng, 0.001);
                    if (patchAreaLatLngMax) {
                        if (patchAreaLatLngMax.length == 0) {
                            patchAreaLatLngMax = me.findRoundPoints(patchAreaLatLng, 0.01);
                        }
                        if (patchAreaLatLngMax.length == 0) {
                            patchAreaLatLngMax = me.findRoundPoints(patchAreaLatLng, 1);
                        }
                        if (patchAreaLatLngMax.length == 0) {
                            patchAreaLatLngMax = patchAreaLatLng;
                        }
                    }

                    var pointSelect = null;
                    var pointSelect = patchAreaLatLngMax[0];

                    for (var n = 0; n < patchAreaLatLngMax.length; n++) {
                        let item = patchAreaLatLngMax[n];
                        if (item[1] > pointSelect[1]) {
                            pointSelect = item;
                        }
                    }

                    me.createVertexArea(pointSelect);
                }

            }, { once: true });
        }

        if (this.isDraw) {
            me.map.container.events.add('mousedown', (e) => {
                e = window.event || e;
                if (e.button == 2) { return; }
                createArea();
            });
        }
        else {
            me._redrawArea();
        }
    };


    DrawingManager.prototype.geodesicArea = function (latLngs) {
        var pointsCount = latLngs.length,
            area = 0.0,
            d2r = Math.PI / 180,
            p1, p2;

        if (pointsCount > 2) {
            for (var i = 0; i < pointsCount; i++) {
                p1 = latLngs[i];
                p2 = latLngs[(i + 1) % pointsCount];
                area += ((p2[1] - p1[1]) * d2r) *
                    (2 + Math.sin(p1[0] * d2r) + Math.sin(p2[0] * d2r));
            }
            area = area * 6378137.0 * 6378137.0 / 2.0;
        }

        return Math.abs(area);
    },

        DrawingManager.prototype.createVertexArea = function (pointSelect) {
            const me = this;
            const geo = {
                geometry: {
                    type: "Point",
                    coordinates: pointSelect
                }
            }

            options = {
                iconLayout: 'default#image',
                iconImageSize: [24, 24],
                iconImageOffset: [-20, -5],
                iconImageHref: me.areaOptions.areaIconVertex,
                draggable: false,
                id: me.vertexCount,
            }

            _vertex = new ymaps.GeoObject(geo, options);
            me.map.geoObjects.add(_vertex);

            _vertex.events.add('click', function (e) {
                if (me.polygonArea.get(e.get('target').options.get('id'))) {
                    let id = me.polygonArea.get(e.get('target').options.get('id'));
                    me.map.geoObjects.remove(e.get('target'));
                    me.map.geoObjects.remove(id);
                    me.polygonArea.delete(id);
                    me.map.events.fire('draw:area_remove', me._convertCoordinates());
                }
            });

            me.vertex.push(_vertex);

        },

        DrawingManager.prototype.findRoundPoints = function (patch, round) {
            const me = this;
            var center = ymaps.util.bounds.getCenter(me._area.geometry.getBounds());
            var result = [];
            var roundMinLat = center[0] - round;
            var roundMaxLat = center[0] + round;
            for (var n = 0; n < patch.length; n++) {
                let item = patch[n];
                if (item[0] >= roundMinLat && item[0] <= roundMaxLat) {
                    if (item[1] > center[1]) {
                        result.push(item);
                    }
                }
            }

            return result;
        },

        DrawingManager.prototype._redrawArea = function () {
            const me = this;
            const latlngs = me._convertlatlngs();

            const options = {
                fillColor: me.areaOptions.fillColor,
                fillOpacity: me.areaOptions.fillOpacity,
                strokeColor: me.areaOptions.strokeColor,
                strokeOpacity: me.areaOptions.strokeOpacity,
                strokeWidth: me.areaOptions.strokeWeight,
                syncOverlayInit: true,
            }

            me._area = new ymaps.Polygon([latlngs,], {}, options);
            me.map.geoObjects.add(me._area);
            me._fitBoundsArea(latlngs);

        };

    DrawingManager.prototype._setIconMarker = function (iconUrl) {

        const me = this;
        if (me._centerMarker) {
            me._centerMarker.options.set({
                iconImageHref: iconUrl,
                iconImageSize: [24, 30],
                iconImageOffset: [-10, -24],
            });
        }
    }

    DrawingManager.prototype._convertlatlngs = function () {

        const me = this;
        let latlngs = [];

        if (me.areaCoordinates.length > 0) {
            for (var n = 0; n < me.areaCoordinates.length; n++) {
                let item = me.areaCoordinates[n];
                let position = [item.latitude, item.longitude];
                latlngs.push(position);
            }

            return latlngs;
        }

    }

    DrawingManager.prototype._reset = function () {
        const me = this;
        me._removeArea();
        me._removeCenterMarker();
        me._removeCircle();
    }

    DrawingManager.prototype._setDrawing = function (enabled) {
        const me = this;
        enabled ? me.map.behaviors.get('dblClickZoom').enable() : me.map.behaviors.get('dblClickZoom').disable();
        enabled ? me.map.behaviors.get('scrollZoom').enable() : me.map.behaviors.get('scrollZoom').disable();
        enabled ? me.map.behaviors.get('drag').enable() : me.map.behaviors.get('drag').disable();
    };

    DrawingManager.prototype._fitBoundsArea = function (latlng) {
        const me = this;
        if (me._area) {
            me.map.setBounds(me._area.geometry.getBounds(), { checkZoomRange: true });
        }
        me.map.events.fire('draw:area_create', me._convertCoordinates(latlng));
    };

    DrawingManager.prototype._fitBoundsCircle = function () {
        const me = this;
        if (!me._circle) { return; }
        me.map.setBounds(me._circle.geometry.getBounds());
        me.map.events.fire('draw:circle_create', me._getInfo());
    };

    DrawingManager.prototype._fitBoundsMarker = function () {
        const me = this;
        if (!me._centerMarker) { return; }
        me.map.getZoom() > 9 ?
            me.map.setCenter(me._centerMarker.geometry.getCoordinates(), me.map.getZoom(), {
                checkZoomRange: true,
            })
            : me.map.setCenter(me._centerMarker.geometry.getCoordinates(), 9, {
                checkZoomRange: true,
            });
    };

    DrawingManager.prototype._convertCoordinates = function (coordinates) {
        let me = this;
        let positions = [];
        let position = null;
        let arrayArea = {};
        let id = null;
        if (!me.isMoreArea) {
            let latLng = [];
            for (var n = 0; n < coordinates.length; n++) {
                let item = coordinates[n];
                let position = {
                    latitude: item[0],
                    longitude: item[1],
                }
                latLng.push(position);
                position = positions;
                id = 0;
            }
            positions.push(latLng);
            position = positions[0];
            let center = {
                latitude: ymaps.util.bounds.getCenter(me._area.geometry.getBounds())[0],
                longitude: ymaps.util.bounds.getCenter(me._area.geometry.getBounds())[1],
            }

            eventArea = {
                id: id,
                positions,
                position,
                center: center,
                zoom: me.map.getZoom(),
            }
            return { eventArea };

        } else {
            me.polygonArea.forEach((key, value) => {
                let latLng = [];
                for (var n = 0; n < key.geometry.getCoordinates().length; n++) {
                    let item = key.geometry.getCoordinates()[n];
                    for (var i = 0; i < item.length; i++) {
                        let it = item[i];
                        let lnlt = {
                            latitude: it[0],
                            longitude: it[1],
                        }
                        latLng.push(lnlt);
                        id = value;
                    }
                    positions.push(latLng);
                }
            });

            eventArea = {
                id: id,
                positions,
            }
            return { eventArea };
        }
    };

    DrawingManager.prototype._createVertexMarker = function () {
        let me = this;
        me.to = null;
        me.to = me.destination(me._centerMarker.geometry.getCoordinates(), 90, me.radius);

        const g = {
            geometry: {
                type: "Point",
                coordinates: me.to
            }
        }

        const svg = [
            '<div style="cursor:col-resize; z-index: 3004;">',
            '<?xml version="1.0"?>',
            '<svg width="15px" height="15px" viewBox="0 0 100 100" version="1.1" xmlns="http://www.w3.org/2000/svg">',
            '<circle stroke="#003dd9" fill="white" stroke-width="10" cx="50" cy="50" r="35"/>',
            '</svg>',
            '</div>'
        ].join('\n');

        const opt = {
            iconLayout: 'default#imageWithContent',
            iconImageSize: [0, 0],
            iconImageOffset: [-9, -5],
            syncOverlayInit: true,
            draggable: true,
            iconContentLayout: ymaps.templateLayoutFactory.createClass(svg),

        }
        me._vertexMarker = new ymaps.GeoObject(g, opt);
        me.map.geoObjects.add(me._vertexMarker);
        me._vertexMarker.options.set('pane', 'vertex');
        me.map.panes.get('vertex').getElement().style.cursor = 'col-resize';
        me._centerMarker && me._vertexMarkerAddEventListener();

    };

    DrawingManager.prototype._createVertexMarkerCopy = function () {

        let me = this;
        const svg = [
            '<?xml version="1.0"?>',
            '<svg width="15px" height="15px" viewBox="0 0 100 100" version="1.1" xmlns="http://www.w3.org/2000/svg">',
            '<circle stroke="#003dd9" fill="white" stroke-width="10" cx="50" cy="50" r="35"/>',
            '</svg>'
        ].join('\n');

        const opt = {
            iconLayout: 'default#imageWithContent',
            iconImageSize: [0, 0],
            iconImageOffset: [-9, -5],
            syncOverlayInit: true,
            draggable: true,
            iconContentLayout: ymaps.templateLayoutFactory.createClass(svg),

        }

        const g = {
            geometry: {
                type: "Point",
                coordinates: me._vertexMarker.geometry.getCoordinates()
            }
        }

        me._vertexMarkerCopy = new ymaps.GeoObject(g, opt);
        me.map.geoObjects.add(me._vertexMarkerCopy);
        me._vertexMarkerCopyAddEventListener();

    };

    DrawingManager.prototype._vertexMarkerCopyAddEventListener = function () {

        me._vertexMarkerCopy.events.add('mouseenter', () => {
            me._vertexMarkerCopy.options.set({
                zIndexHover: 100,
                zIndex: 100,
                zIndexDrag: 100,
                visible: false,
            });

            me._vertexMarker.options.set({
                visible: true,
                zIndexHover: 200,
                zIndex: 200,
                zIndexDrag: 200
            });
        });
    };

    DrawingManager.prototype._vertexMarkerAddEventListener = function () {

        const me = this;
        me.isMaxDistance = false;
        me._vertexMarker.events.add('drag', (event) => {
            if (!me._centerMarker) { return; }
            me.isMaxDistance = false;
            let coords = event.get('target').geometry.getCoordinates();
            let distance = me.getDistanceTo(me._centerMarker.geometry.getCoordinates(), coords);
            let offset = me.map.container.getOffset();

            if (distance >= me.circleOptions.maxRadius) {
                me._vertexMarker.options.set({
                    visible: false,
                    zIndexHover: 100,
                    zIndex: 100,
                    zIndexDrag: 100,
                });

                if (me._vertexMarkerCopy) {
                    me._vertexMarkerCopy.options.set({
                        visible: true,
                        zIndexHover: 200,
                        zIndex: 200,
                        zIndexDrag: 200,
                    });
                }

                me.isMaxDistance = true;
                me.map.panes.get('events').getElement().style.cursor = 'not-allowed';
                me.map.events.fire('draw:circle_radius_max', null);
                return false;
            }

            me.radius = distance;
            me._circle && me._circle.geometry.setRadius(distance);
            me._vertexMarkerCopy && me.map.geoObjects.remove(me._vertexMarkerCopy);
            !me.isMaxDistance && me._createVertexMarkerCopy();

            me._vertexMarker.options.set({
                zIndexHover: 100,
                zIndex: 100,
                zIndexDrag: 100,
                visible: true,
            });

            if (me._vertexMarkerCopy) {
                me._vertexMarkerCopy.options.set({
                    visible: false,
                    zIndexHover: 200,
                    zIndexDrag: 200,
                    zIndex: 200
                });
            }

            const pixel = {
                clientX: event.get('domEvent').get('clientX') - offset[0],
                clientY: event.get('domEvent').get('clientY') - offset[1],
            }

            let eventCircle = {
                pixel,
                isDraw: me.isDraw,
                radius: Math.floor(me._circle.geometry.getRadius()),
            }

            me._vertexMarker.events.add('mouseenter', () => {
                me._vertexMarkerCopy.options.set({
                    zIndexHover: 100,
                    zIndex: 100,
                    zIndexDrag: 100,
                    visible: false,
                });

                me._vertexMarker.options.set({
                    visible: true,
                    zIndexHover: 200,
                    zIndex: 200,
                    zIndexDrag: 200
                });

            });

            me.map.events.fire('draw:circle_radius_change', { eventCircle });

        });

        me._vertexMarker.events.add('dragend', () => {

            me.map.panes.get('events').getElement().style.cursor = 'inherit';
            if (me._vertexMarkerCopy) {

                if (me.isMaxDistance) {
                    me._vertexMarker.geometry.setCoordinates(me._vertexMarkerCopy.geometry.getCoordinates());
                    me._vertexMarker.options.set({
                        zIndexHover: 100,
                        zIndex: 100,
                        zIndexDrag: 100,
                        visible: true,
                    });
                }
            }

            me.map.events.fire('draw:circle_radius_complete', me._getInfo());
        });
    };

    DrawingManager.prototype._centerMarkerAddEventListener = function () {

        const me = this;
        me._centerMarker.events.add('drag', function (event) {

            event.get('target').getOverlay().then((e) => {
                e.options.set('cursor', 'move')
            })

            if (me._vertexMarkerCopy) {
                me._vertexMarkerCopy.options.set({
                    visible: false,
                    zIndexHover: 100,
                    zIndex: 100,
                    zIndexDrag: 100
                });
            }

            me._vertexMarker.options.set({
                visible: false,
                zIndexHover: 200,
                zIndex: 200,
                zIndexDrag: 200
            });

            let coords = event.get('target').geometry.getCoordinates();


            me._circle.geometry.setCoordinates(coords);
            let to = me.destination(coords, 90, me.radius);

            me._vertexMarker && me._vertexMarker.geometry.setCoordinates(to);
            me._vertexMarkerCopy && me._vertexMarkerCopy.geometry.setCoordinates(to);


            me.map.events.fire('draw:circle_centre_change', me._getInfo());
        })

        me._centerMarker.events.add('dragend', () => {
            me._vertexMarker.options.set({
                visible: true,
            });

            me.map.events.fire('draw:circle_center_complete', me._getInfo());
        });

    };

    DrawingManager.prototype._removeCircle = function () {
        const me = this;
        if (me._circle) {
            me.map.geoObjects.remove(me._circle);
            me._circle = null;
            me.map.events.fire('draw:circle_remove', null);
            if (me._vertexMarker) {
                me.map.geoObjects.remove(me._vertexMarker);
                me._vertexMarker = null;
            }

            if (me._vertexMarkerCopy) {
                me.map.geoObjects.remove(me._vertexMarkerCopy);
                me._vertexMarkerCopy = null;
            }
            me._circle = null;
        }
    };

    DrawingManager.prototype._removeCenterMarker = function () {
        const me = this;
        if (me._centerMarker && !me._circle) {
            me.map.geoObjects.remove(me._centerMarker);
            me._centerMarker = null;
            me.map.events.fire('draw:marker_remove', null);
        }
    };

    DrawingManager.prototype._removeArea = function () {
        let me = this;
        if (me._area) {
            me.map.geoObjects.remove(me._area);
            me._area = null;
            me.map.events.fire('draw:area_remove', []);
        }

        me.polygonArea.forEach((key) => {
            me.map.geoObjects.remove(key);
        });

        me.vertex.forEach((key) => {
            me.map.geoObjects.remove(key);
        });

        me.polygonArea.clear();
        me.vertex = [];
        me.currentAreaCount = 0;
    };

    DrawingManager.prototype._getInfo = function () {
        const me = this;
        const position = {
            latitude: me._centerMarker && me._centerMarker.geometry.getCoordinates()[0] || me._circle.geometry.getCoordinates()[0],
            longitude: me._centerMarker && me._centerMarker.geometry.getCoordinates()[1] || me._circle.geometry.getCoordinates()[1]
        }
        const eventCircle = {
            isDraw: me.isDraw,
            bounds: me._circle.geometry.getBounds(),
            radius: Math.floor(me._circle.geometry.getRadius()),
            position,
            zoom: me.map.getZoom(),
        };

        return { eventCircle };
    };

    DrawingManager.prototype._getPosition = function () {
        let me = this;
        let position = {
            latitude: me._centerMarker.geometry.getCoordinates()[0],
            longitude: me._centerMarker.geometry.getCoordinates()[1]
        }
        let eventMarker = {
            isDraw: me.isDraw,
            position,
            placeId: null,
        };

        return { eventMarker };
    };

    DrawingManager.prototype._getZoom = function () {
        let me = this;
        let zoom = {
            zoom: me.map.getZoom()
        }
        return zoom;
    };

    DrawingManager.prototype.destination = function (latlng, heading, distance) {
        heading = (heading + 360) % 360;
        var rad = Math.PI / 180,
            radInv = 180 / Math.PI,
            R = 6378137,
            lon1 = latlng[1] * rad,
            lat1 = latlng[0] * rad,
            rheading = heading * rad,
            sinLat1 = Math.sin(lat1),
            cosLat1 = Math.cos(lat1),
            cosDistR = Math.cos(distance / R),
            sinDistR = Math.sin(distance / R),
            lat2 = Math.asin(sinLat1 * cosDistR + cosLat1 *
                sinDistR * Math.cos(rheading)),
            lon2 = lon1 + Math.atan2(Math.sin(rheading) * sinDistR *
                cosLat1, cosDistR - sinLat1 * Math.sin(lat2));
        lon2 = lon2 * radInv;
        lon2 = lon2 > 180 ? lon2 - 360 : lon2 < -180 ? lon2 + 360 : lon2;
        return [lat2 * radInv, lon2]

    };

    DrawingManager.prototype.degreeToRad = function (degree) {
        return Math.PI * degree / 180;
    };

    DrawingManager.prototype._getRange = function (v, a, b) {
        if (a != null) {
            v = Math.max(v, a);
        }
        if (b != null) {
            v = Math.min(v, b);
        }
        return v;
    };

    DrawingManager.prototype._getLoop = function (v, a, b) {
        while (v > b) {
            v -= b - a
        }
        while (v < a) {
            v += b - a
        }
        return v;
    };

    DrawingManager.prototype.getDistanceTo = function (point1, point2) {
        let me = this;
        point1[1] = me._getLoop(point1[1], -180, 180);
        point1[0] = me._getRange(point1[0], -74, 74);
        point2[1] = me._getLoop(point2[1], -180, 180);
        point2[0] = me._getRange(point2[0], -74, 74);
        var x1, x2, y1, y2;
        x1 = this.degreeToRad(point1[1]);
        y1 = this.degreeToRad(point1[0]);
        x2 = this.degreeToRad(point2[1]);
        y2 = this.degreeToRad(point2[0]);
        return 6370996.81 * Math.acos((Math.sin(y1) * Math.sin(y2) + Math.cos(y1) * Math.cos(y2) * Math.cos(x2 - x1)));
    };

    DrawingManager.prototype.extend = function (obj1, obj2) {
        return (function (object) {
            var property;
            for (property in object.prototype) {
                this.prototype[property] = object.prototype[property];
            }
            return this;
        }).apply(obj1, [obj2]);
    };
};

if (typeof module == 'object') {
    module.exports = { default: DrawingManager, DrawingManager: DrawingManager };
}
